{# Question Presentation #}
{{question_text}}
{% if use_code %}
    {%- for option in question_options %}
{{ loop.index0 }}: {{option}}
    {% endfor %}
{% else %}
    {% for option in question_options %}
{{option}}
    {% endfor %}
{% endif %}
{% if other_instructions %}
{{ other_instructions }}
{% else %}
If none of the provided options apply, select "Other" and provide your specific response.
{% endif %}
Only 1 option may be selected.