{{question_text}}

Your response should be only a valid JSON in the following format:
{% if include_comment %}
    {
    "answer": [<comma-separated list of responsive words or phrases as independent strings>],
    "comment": "<put comment here>"
    }
{% else %}
    {
    "answer": [<comma-separated list of responsive words or phrases as independent strings>],
    }
{% endif %}

{% if min_list_items is not none %}
The list must contain at least {{ min_list_items }} items.
{% endif %}
{% if max_list_items is not none %}
The list must not contain more than {{ max_list_items }} items.
{% endif %}
